# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Neuron classes and functions."""

from collections import deque
from itertools import chain

import morphio
import numpy as np
from neurom import morphmath
from neurom.core.soma import make_soma
from neurom.core.dataformat import COLS
from neurom.core.types import NeuriteIter, NeuriteType


class Section:
    """Simple recursive tree class."""

    def __init__(self, morphio_section):
        """The section constructor."""
        self.morphio_section = morphio_section

    @property
    def id(self):
        """Returns the section ID."""
        return self.morphio_section.id

    @property
    def parent(self):
        """Returns the parent section if non root section else None."""
        if self.morphio_section.is_root:
            return None
        return Section(self.morphio_section.parent)

    @property
    def children(self):
        """Returns a list of child section."""
        return [Section(child) for child in self.morphio_section.children]

    def append_section(self, section):
        """Appends a section to the current section object.

        Args:
            section (morphio.Section|morphio.mut.Section|Section|morphio.PointLevel): a section
        """
        if isinstance(section, Section):
            return self.morphio_section.append_section(section.morphio_section)
        return self.morphio_section.append_section(section)

    def is_forking_point(self):
        """Is this section a forking point?"""
        return len(self.children) > 1

    def is_bifurcation_point(self):
        """Is tree a bifurcation point?"""
        return len(self.children) == 2

    def is_leaf(self):
        """Is tree a leaf?"""
        return len(self.children) == 0

    def is_root(self):
        """Is tree the root node?"""
        return self.parent is None

    def ipreorder(self):
        """Depth-first pre-order iteration of tree nodes."""
        children = deque((self, ))
        while children:
            cur_node = children.pop()
            children.extend(reversed(cur_node.children))
            yield cur_node

    def ipostorder(self):
        """Depth-first post-order iteration of tree nodes."""
        children = [self, ]
        seen = set()
        while children:
            cur_node = children[-1]
            if cur_node not in seen:
                seen.add(cur_node)
                children.extend(reversed(cur_node.children))
            else:
                children.pop()
                yield cur_node

    def iupstream(self):
        """Iterate from a tree node to the root nodes."""
        t = self
        while t is not None:
            yield t
            t = t.parent

    def ileaf(self):
        """Iterator to all leaves of a tree."""
        return filter(Section.is_leaf, self.ipreorder())

    def iforking_point(self, iter_mode=ipreorder):
        """Iterator to forking points.

        Args:
            iter_mode: iteration mode. Default: ipreorder.
        """
        return filter(Section.is_forking_point, iter_mode(self))

    def ibifurcation_point(self, iter_mode=ipreorder):
        """Iterator to bifurcation points.

        Args:
            iter_mode: iteration mode. Default: ipreorder.
        """
        return filter(Section.is_bifurcation_point, iter_mode(self))

    def __eq__(self, other):
        """Equal when its morphio section is equal."""
        return self.morphio_section == other.morphio_section

    def __hash__(self):
        """Hash of its id."""
        return self.id

    def __nonzero__(self):
        """If has children."""
        return self.morphio_section is not None

    __bool__ = __nonzero__

    @property
    def points(self):
        """Returns the section list of points the NeuroM way (points + radius)."""
        return np.concatenate((self.morphio_section.points,
                               self.morphio_section.diameters[:, np.newaxis] / 2.),
                              axis=1)

    @points.setter
    def points(self, value):
        """Set the points."""
        self.morphio_section.points = np.copy(value[:, COLS.XYZ])
        self.morphio_section.diameters = np.copy(value[:, COLS.R]) * 2

    @property
    def type(self):
        """Returns the section type."""
        return NeuriteType(int(self.morphio_section.type))

    @property
    def length(self):
        """Return the path length of this section."""
        return morphmath.section_length(self.points)

    @property
    def area(self):
        """Return the surface area of this section.

        The area is calculated from the segments, as defined by this
        section's points
        """
        return sum(morphmath.segment_area(s) for s in iter_segments(self))

    @property
    def volume(self):
        """Return the volume of this section.

        The volume is calculated from the segments, as defined by this
        section's points
        """
        return sum(morphmath.segment_volume(s) for s in iter_segments(self))

    def __repr__(self):
        """Text representation."""
        parent_id = None if self.parent is None else self.parent.id
        return (f'Section(id={self.id}, type={self.type}, n_points={len(self.points)})'
                f'<parent: Section(id={parent_id}), nchildren: {len(self.children)}>')


# NRN simulator iteration order
# See:
# https://github.com/neuronsimulator/nrn/blob/2dbf2ebf95f1f8e5a9f0565272c18b1c87b2e54c/share/lib/hoc/import3d/import3d_gui.hoc#L874
NRN_ORDER = {NeuriteType.soma: 0,
             NeuriteType.axon: 1,
             NeuriteType.basal_dendrite: 2,
             NeuriteType.apical_dendrite: 3,
             NeuriteType.undefined: 4}


def iter_neurites(obj, mapfun=None, filt=None, neurite_order=NeuriteIter.FileOrder):
    """Iterator to a neurite, neuron or neuron population.

    Applies optional neurite filter and mapping functions.

    Arguments:
        obj: a neurite, neuron or neuron population.
        mapfun: optional neurite mapping function.
        filt: optional neurite filter function.
        neurite_order (NeuriteIter): order upon which neurites should be iterated
            - NeuriteIter.FileOrder: order of appearance in the file
            - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical

    Examples:
        Get the number of points in each neurite in a neuron population

        >>> from neurom.core.neuron import iter_neurites
        >>> n_points = [n for n in iter_neurites(pop, lambda x : len(x.points))]

        Get the number of points in each axon in a neuron population

        >>> import neurom as nm
        >>> from neurom.core.neuron import iter_neurites
        >>> filter = lambda n : n.type == nm.AXON
        >>> mapping = lambda n : len(n.points)
        >>> n_points = [n for n in iter_neurites(pop, mapping, filter)]
    """
    neurites = ((obj,) if isinstance(obj, Neurite) else
                obj.neurites if hasattr(obj, 'neurites') else obj)
    if neurite_order == NeuriteIter.NRN:
        last_position = max(NRN_ORDER.values()) + 1
        neurites = sorted(neurites, key=lambda neurite: NRN_ORDER.get(neurite.type, last_position))

    neurite_iter = iter(neurites) if filt is None else filter(filt, neurites)
    return neurite_iter if mapfun is None else map(mapfun, neurite_iter)


def iter_sections(neurites,
                  iterator_type=Section.ipreorder,
                  neurite_filter=None,
                  neurite_order=NeuriteIter.FileOrder):
    """Iterator to the sections in a neurite, neuron or neuron population.

    Arguments:
        neurites: neuron, population, neurite, or iterable containing neurite objects
        iterator_type: section iteration order within a given neurite. Must be one of:
            Section.ipreorder: Depth-first pre-order iteration of tree nodes
            Section.ipostorder: Depth-first post-order iteration of tree nodes
            Section.iupstream: Iterate from a tree node to the root nodes
            Section.ibifurcation_point: Iterator to bifurcation points
            Section.ileaf: Iterator to all leaves of a tree

        neurite_filter: optional top level filter on properties of neurite neurite objects.
        neurite_order (NeuriteIter): order upon which neurites should be iterated
            - NeuriteIter.FileOrder: order of appearance in the file
            - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical


    Examples:
        Get the number of points in each section of all the axons in a neuron population

        >>> import neurom as nm
        >>> from neurom.core.neuron import iter_sections
        >>> filter = lambda n : n.type == nm.AXON
        >>> n_points = [len(s.points) for s in iter_sections(pop,  neurite_filter=filter)]
    """
    return chain.from_iterable(
        iterator_type(neurite.root_node) for neurite in
        iter_neurites(neurites, filt=neurite_filter, neurite_order=neurite_order))


def iter_segments(obj, neurite_filter=None, neurite_order=NeuriteIter.FileOrder):
    """Return an iterator to the segments in a collection of neurites.

    Arguments:
        obj: neuron, population, neurite, section, or iterable containing neurite objects
        neurite_filter: optional top level filter on properties of neurite neurite objects
        neurite_order: order upon which neurite should be iterated. Values:
            - NeuriteIter.FileOrder: order of appearance in the file
            - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical

    Note:
        This is a convenience function provided for generic access to
        neuron segments. It may have a performance overhead WRT custom-made
        segment analysis functions that leverage numpy and section-wise iteration.
    """
    sections = iter((obj,) if isinstance(obj, Section) else
                    iter_sections(obj,
                                  neurite_filter=neurite_filter,
                                  neurite_order=neurite_order))

    return chain.from_iterable(zip(sec.points[:-1], sec.points[1:])
                               for sec in sections)


def graft_neuron(section):
    """Returns a neuron starting at section."""
    assert isinstance(section, Section)
    m = morphio.mut.Morphology()
    m.append_root_section(section.morphio_section)
    return Neuron(m)


class Neurite:
    """Class representing a neurite tree."""

    def __init__(self, root_node):
        """Constructor.

        Args:
            root_node (morphio.Section): root section
        """
        self.morphio_root_node = root_node

    @property
    def root_node(self):
        """The first section of the neurite."""
        return Section(self.morphio_root_node)

    @property
    def type(self):
        """The type of the root node."""
        return self.root_node.type

    @property
    def points(self):
        """Array with all the points in this neurite.

        Note: Duplicate points at section bifurcations are removed
        """
        # Neurite first point must be added manually
        _ptr = [self.root_node.points[0][COLS.XYZR]]
        for s in iter_sections(self):
            _ptr.append(s.points[1:, COLS.XYZR])
        return np.vstack(_ptr)

    @property
    def length(self):
        """Returns the total length of this neurite.

        The length is defined as the sum of lengths of the sections.
        """
        return sum(s.length for s in self.iter_sections())

    @property
    def area(self):
        """Return the surface area of this neurite.

        The area is defined as the sum of area of the sections.
        """
        return sum(s.area for s in self.iter_sections())

    @property
    def volume(self):
        """Return the volume of this neurite.

        The volume is defined as the sum of volumes of the sections.
        """
        return sum(s.volume for s in self.iter_sections())

    def iter_sections(self, order=Section.ipreorder, neurite_order=NeuriteIter.FileOrder):
        """Iteration over section nodes.

        Arguments:
            order: section iteration order within a given neurite. Must be one of:
                Section.ipreorder: Depth-first pre-order iteration of tree nodes
                Section.ipreorder: Depth-first post-order iteration of tree nodes
                Section.iupstream: Iterate from a tree node to the root nodes
                Section.ibifurcation_point: Iterator to bifurcation points
                Section.ileaf: Iterator to all leaves of a tree

            neurite_order: order upon which neurites should be iterated. Values:
                - NeuriteIter.FileOrder: order of appearance in the file
                - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical
        """
        return iter_sections(self, iterator_type=order, neurite_order=neurite_order)

    def __nonzero__(self):
        """If has root node."""
        return bool(self.morphio_root_node)

    def __eq__(self, other):
        """If root node ids and types are equal."""
        return self.type == other.type and self.morphio_root_node.id == other.morphio_root_node.id

    def __hash__(self):
        """Hash is made of tuple of type and root_node."""
        return hash((self.type, self.root_node))

    __bool__ = __nonzero__

    def __repr__(self):
        """Return a string representation."""
        return 'Neurite <type: %s>' % self.type


class Neuron(morphio.mut.Morphology):
    """Class representing a simple neuron."""

    def __init__(self, filename, name=None):
        """Neuron constructor.

        Args:
            filename (str|Path): a filename
            name (str): a option neuron name
        """
        try:
            morphio.set_ignored_warning([morphio.Warning.wrong_root_point], True)
            morphio.set_raise_warnings(True)
            super().__init__(filename)
        finally:
            morphio.set_raise_warnings(False)
        self.name = name if name else 'Neuron'
        self.morphio_soma = super().soma
        self.neurom_soma = make_soma(self.morphio_soma)

    @property
    def soma(self):
        """Corresponding soma."""
        return self.neurom_soma

    @property
    def neurites(self):
        """The list of neurites."""
        return [Neurite(root_section) for root_section in self.root_sections]

    @property
    def sections(self):
        """The array of all sections, excluding the soma."""
        return list(iter_sections(self))

    @property
    def points(self):
        """Returns the list of points."""
        return np.concatenate(
            [section.points for section in iter_sections(self)])

    def transform(self, trans):
        """Return a copy of this neuron with a 3D transformation applied."""
        obj = Neuron(self)
        obj.morphio_soma.points = trans(obj.morphio_soma.points)

        for section in obj.sections:
            section.morphio_section.points = trans(section.morphio_section.points)
        return obj

    def __copy__(self):
        """Creates a deep copy of Neuron instance."""
        return Neuron(self, self.name)

    def __deepcopy__(self, memodict={}):
        """Creates a deep copy of Neuron instance."""
        # pylint: disable=dangerous-default-value
        return Neuron(self, self.name)

    def __repr__(self):
        """Return a string representation."""
        return 'Neuron <soma: %s, n_neurites: %d>' % \
            (self.soma, len(self.neurites))



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path
import neurom as nm
from numpy.testing import assert_array_equal, assert_almost_equal
import numpy as np

SWC_PATH = Path(__file__).parent.parent / 'data/swc/'


def test_section_base_func():
    nrn = nm.load_neuron(str(SWC_PATH / 'simple.swc'))
    section = nrn.sections[0]

    assert section.type == nm.NeuriteType.basal_dendrite
    assert section.id == 0
    assert_array_equal(section.points, [[0, 0, 0, 1], [0, 5, 0, 1]])
    assert_almost_equal(section.length, 5)
    assert_almost_equal(section.area, 31.41592653589793)
    assert_almost_equal(section.volume, 15.707963267948964)


def test_section_tree():
    nrn = nm.load_neuron(str(SWC_PATH / 'simple.swc'))

    assert nrn.sections[0].parent is None
    assert nrn.sections[0] == nrn.sections[0].children[0].parent

    assert_array_equal([s.is_root() for s in nrn.sections],
                       [True, False, False, True, False, False])
    assert_array_equal([s.is_leaf() for s in nrn.sections],
                       [False, True, True, False, True, True])
    assert_array_equal([s.is_forking_point() for s in nrn.sections],
                       [True, False, False, True, False, False])
    assert_array_equal([s.is_bifurcation_point() for s in nrn.sections],
                       [True, False, False, True, False, False])
    assert_array_equal([s.id for s in nrn.neurites[0].root_node.ipreorder()],
                       [0, 1, 2])
    assert_array_equal([s.id for s in nrn.neurites[0].root_node.ipostorder()],
                       [1, 2, 0])
    assert_array_equal([s.id for s in nrn.neurites[0].root_node.iupstream()],
                       [0])
    assert_array_equal([s.id for s in nrn.sections[2].iupstream()],
                       [2, 0])
    assert_array_equal([s.id for s in nrn.neurites[0].root_node.ileaf()],
                       [1, 2])
    assert_array_equal([s.id for s in nrn.sections[2].ileaf()],
                       [2])
    assert_array_equal([s.id for s in nrn.neurites[0].root_node.iforking_point()],
                       [0])
    assert_array_equal([s.id for s in nrn.neurites[0].root_node.ibifurcation_point()],
                       [0])


def test_append_section():
    n = nm.load_neuron(SWC_PATH / 'simple.swc')
    s = n.sections[0]

    s.append_section(n.sections[-1])
    assert len(s.children) == 3
    assert s.children[-1].id == 6
    assert s.children[-1].type == n.sections[-1].type

    s.append_section(n.sections[-1].morphio_section)
    assert len(s.children) == 4
    assert s.children[-1].id == 7
    assert s.children[-1].type == n.sections[-1].type


def test_set_points():
    n = nm.load_neuron(SWC_PATH / 'simple.swc')
    s = n.sections[0]
    s.points = np.array([
        [0, 5, 0, 2],
        [0, 7, 0, 2],
    ])
    assert_array_equal(s.points, np.array([
        [0, 5, 0, 2],
        [0, 7, 0, 2],
    ]))



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path

import neurom as nm
import numpy as np
from neurom import geom
from numpy.testing import assert_almost_equal

SWC_DATA_PATH = Path(__file__).parent.parent / 'data/swc'
NRN = nm.load_neuron(SWC_DATA_PATH / 'Neuron.swc')
SIMPLE = nm.load_neuron(SWC_DATA_PATH / 'simple.swc')

class PointObj:
    pass


def test_bounding_box():

    pts = np.array([[-1, -2, -3, -999],
                    [1, 2, 3, 1000],
                    [-100, 5, 33, 42],
                    [42, 55, 12, -3]])

    obj = PointObj()
    obj.points = pts

    assert np.alltrue(geom.bounding_box(obj) == [[-100, -2, -3], [42, 55, 33]])


def test_bounding_box_neuron():

    ref = np.array([[-40.32853516, -57.600172, 0.],
                    [64.74726272, 48.51626225, 54.20408797]])

    assert np.allclose(geom.bounding_box(NRN), ref)


def test_bounding_box_soma():
    ref = np.array([[0., 0., 0.], [0.1, 0.2, 0.]])
    assert np.allclose(geom.bounding_box(NRN.soma), ref)


def test_bounding_box_neurite():
    nrt = SIMPLE.neurites[0]
    ref = np.array([[-5.,  0.,  0.], [ 6.,  5.,  0.]])
    np.testing.assert_allclose(geom.bounding_box(nrt), ref)

def test_convex_hull_points():

    # This leverages scipy ConvexHull and we don't want
    # to re-test scipy, so simply check that the points are the same.
    hull = geom.convex_hull(NRN)
    assert np.alltrue(hull.points == NRN.points[:, :3])


def test_convex_hull_volume():

    # This leverages scipy ConvexHull and we don't want
    # to re-test scipy, so simply regression test the volume
    hull = geom.convex_hull(NRN)
    assert_almost_equal(hull.volume, 208641, decimal=0)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
import warnings
from io import StringIO

import numpy as np
from morphio import MorphioError, SomaError
from neurom import load_neuron
from neurom.core import soma
from mock import Mock

import pytest
from numpy.testing import assert_array_equal, assert_almost_equal


def test_no_soma_builder():
    morphio_soma = Mock()
    morphio_soma.type = None
    with pytest.raises(SomaError, match='No NeuroM constructor for MorphIO soma type'):
        soma.make_soma(morphio_soma)


def test_no_soma():
    sm = load_neuron(StringIO(u"""
        ((Dendrite)
        (0 0 0 1.0)
        (0 0 0 2.0))"""), reader='asc').soma
    assert sm.center is None
    assert sm.points.shape == (0, 4)


def test_Soma_SinglePoint():
    sm = load_neuron(StringIO(u"""1 1 11 22 33 44 -1"""), reader='swc').soma
    assert 'SomaSinglePoint' in str(sm)
    assert isinstance(sm, soma.SomaSinglePoint)
    assert list(sm.center) == [11, 22, 33]
    assert sm.radius == 44


def test_Soma_contour():
    with warnings.catch_warnings(record=True):
        sm = load_neuron(StringIO(u"""((CellBody)
                                      (0 0 0 44)
                                      (0 -44 0 44)
                                      (0 +44 0 44))"""), reader='asc').soma

    assert 'SomaSimpleContour' in str(sm)
    assert isinstance(sm, soma.SomaSimpleContour)
    assert list(sm.center) == [0, 0, 0]
    assert_almost_equal(sm.radius, 29.33333333, decimal=6)


def test_Soma_ThreePointCylinder():
    sm = load_neuron(StringIO(u"""1 1 0   0 0 44 -1
                                  2 1 0 -44 0 44  1
                                  3 1 0 +44 0 44  1"""), reader='swc').soma
    assert 'SomaNeuromorphoThreePointCylinders' in str(sm)
    assert isinstance(sm, soma.SomaNeuromorphoThreePointCylinders)
    assert list(sm.center) == [0, 0, 0]
    assert sm.radius == 44


def test_Soma_ThreePointCylinder_invalid_radius():
    with warnings.catch_warnings(record=True) as w_list:
        load_neuron(StringIO(u"""
                        1 1 0   0 0 1e-8 -1
                        2 1 0 -1e-8 0 1e-8  1
                        3 1 0 +1e-8 0 1e-8  1"""), reader='swc').soma
        assert 'Zero radius for SomaNeuromorphoThreePointCylinders' in str(w_list[0])


def test_Soma_ThreePointCylinder_invalid():
    with pytest.raises(MorphioError, match='Warning: the soma does not conform the three point soma spec'):
        load_neuron(StringIO(u"""
                        1 1 0   0 0 1e-8 -1
                        2 1 0 -44 0 1e-8  1
                        3 1 0 +44 0 1e-8  1"""), reader = 'swc')


def check_SomaC(stream):
    sm = load_neuron(StringIO(stream), reader='asc').soma
    assert 'SomaSimpleContour' in str(sm)
    assert isinstance(sm, soma.SomaSimpleContour)
    np.testing.assert_almost_equal(sm.center, [0., 0., 0.])
    assert_almost_equal(sm.radius, 1.0)


def test_SomaC():
    with warnings.catch_warnings(record=True):
        check_SomaC(u"""((CellBody)
                        (1 0 0 44)
                        (0 1 0 44)
                        (-1 0 0 44)
                        (0 -1 0 44)) """)

        sin_pi_by_4 = math.cos(math.pi/4.)
        cos_pi_by_4 = math.sin(math.pi/4.)

        check_SomaC(u"""((CellBody)
                         (1 0 0 44)
                         ({sin} {cos} 0 44)
                         (0 1 0 44)
                         (-1 0 0 44)
                         (-{sin} -{cos} 0 44)
                         (0 -1 0 44))""".format(sin=sin_pi_by_4,
                                                cos=cos_pi_by_4))

        check_SomaC(u"""((CellBody)
        (1 0 0 44)
        ({sin} {cos} 0 44)
        (0 1 0 44)
        (-{sin} {cos} 0 44)
        (-1 0 0 44)
        (-{sin} -{cos} 0 44)
        (0 -1 0 44)
        ({sin} -{cos} 0 44))""".format(sin=sin_pi_by_4,
                                       cos=cos_pi_by_4))


def test_soma_points_2():
    load_neuron(StringIO(u"""
                    1 1 0 0 -10 40 -1
                    2 1 0 0   0 40  1"""), reader='swc').soma
    load_neuron(StringIO(u"""((CellBody)
                             (0 0 0 44)
                             (0 +44 0 44))"""), reader='asc').soma


def test_Soma_Cylinders():
    s = load_neuron(StringIO(u"""
                1 1 0 0 -10 40 -1
                2 1 0 0   0 40  1
                3 1 0 0  10 40  2"""), reader='swc').soma

    # if r = 2*h (ie: as in this case 10 - -10 == 20), then the
    # area of a cylinder (excluding end caps) is:
    # 2*pi*r*h == 4*pi*r^2 == area of a sphere of radius 20
    assert s.radius == 20.0
    assert_almost_equal(s.area, 5026.548245743669)
    assert_array_equal(s.center, [0, 0, -10])
    assert 'SomaCylinders' in str(s)

    # neuromorpho style
    with warnings.catch_warnings(record=True):
        s = load_neuron(StringIO(u"""
                1 1 0   0 0 10 -1
                2 1 0 -10 0 10  1
                3 1 0  10 0 10  1"""), reader='swc').soma

    assert 'SomaNeuromorphoThreePointCylinders' in str(s)
    assert list(s.center) == [0., 0., 0.]
    assert_almost_equal(s.area, 1256.6370614)

    # some neuromorpho files don't follow the convention
    #but have (ys + rs) as point 2, and have xs different in each line
    # ex: http://neuromorpho.org/dableFiles/brumberg/CNG%20version/april11s1cell-1.CNG.swc
    with warnings.catch_warnings(record=True):
        s = load_neuron(StringIO(u"""
                1 1  0  0 0 10 -1
                2 1 -2 -6 0 10  1
                3 1  2  6 0 10  1"""), reader='swc').soma

    assert 'SomaNeuromorphoThreePointCylinders' in str(s)
    assert list(s.center) == [0., 0., 0.]
    assert_almost_equal(s.area, 794.76706126368811, decimal=5)

    s = load_neuron(StringIO(u"""
                1 1 0  0 0  0 -1
                2 1 0  2 0  2  1
                3 1 0  4 0  4  2
                4 1 0  6 0  6  3
                5 1 0  8 0  8  4
                6 1 0 10 0 10  5"""), reader='swc').soma

    assert list(s.center) == [0., 0., 0.]
    assert_almost_equal(s.area, 444.288293851) # cone area, not including bottom



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from io import StringIO
from pathlib import Path

import neurom as nm
from neurom import COLS, load_neuron
from neurom.core.neuron import NeuriteIter, Section, iter_neurites, iter_sections, iter_segments
from neurom.core.population import Population
from numpy.testing import assert_array_equal

DATA_PATH = Path(__file__).parent.parent / 'data'

NRN1 = load_neuron(DATA_PATH / 'swc/Neuron.swc')

NEURONS = [NRN1,
           load_neuron(DATA_PATH / 'swc/Single_basal.swc'),
           load_neuron(DATA_PATH / 'swc/Neuron_small_radius.swc'),
           load_neuron(DATA_PATH / 'swc/Neuron_3_random_walker_branches.swc'),]
TOT_NEURITES = sum(len(N.neurites) for N in NEURONS)

SIMPLE = load_neuron(DATA_PATH / 'swc/simple.swc')
REVERSED_NEURITES = load_neuron(DATA_PATH / 'swc/ordering/reversed_NRN_neurite_order.swc')

POP = Population(NEURONS, name='foo')


def assert_sequence_equal(a, b):
    assert tuple(a) == tuple(b)


def test_iter_neurites_default():
    assert list(POP.neurites) == [n for n in iter_neurites(POP)]


def test_iter_neurites_nrn_order():
    actual = list(iter_neurites(REVERSED_NEURITES, neurite_order=NeuriteIter.NRN))
    expected = list(reversed(list(iter_neurites(REVERSED_NEURITES))))
    assert actual == expected


def test_iter_neurites_filter():

    for ntyp in nm.NEURITE_TYPES:
        a = [n for n in POP.neurites if n.type == ntyp]
        b = [n for n in iter_neurites(POP, filt=lambda n: n.type == ntyp)]
        assert a == b


def test_iter_neurites_mapping():

    n = [n for n in iter_neurites(POP, mapfun=lambda n: len(n.points))]
    ref = [211, 211, 211, 211, 211, 211, 211, 211, 211, 500, 500, 500]
    assert n == ref


def test_iter_neurites_filter_mapping():
    n = [n for n in iter_neurites(POP,
                                       mapfun=lambda n: len(n.points),
                                       filt=lambda n: len(n.points) > 250)]

    ref = [500, 500, 500]
    assert n == ref


def test_iter_sections_default():

    ref = [s for n in POP.neurites for s in n.iter_sections()]
    assert (ref ==
                          [n for n in iter_sections(POP)])

def test_iter_sections_default_pop():
    ref = [s.id for n in POP.neurites for s in n.iter_sections()]
    assert ref == [n.id for n in iter_sections(POP)]


def test_iter_sections_filter():
    for ntyp in nm.NEURITE_TYPES:
        a = [s.id for n in filter(lambda nn: nn.type == ntyp, POP.neurites)
             for s in n.iter_sections()]
        b = [n.id for n in iter_sections(POP, neurite_filter=lambda n: n.type == ntyp)]
        assert a == b

def test_iter_sections_inrnorder():
    assert ([s.id for n in POP.neurites for s in n.iter_sections(neurite_order=NeuriteIter.NRN)] ==
                          [0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83,  0, 1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83,  0,  1,  2])

def test_iter_sections_ipreorder():
    assert ([s.id for n in POP.neurites for s in n.iter_sections(Section.ipreorder)] ==
                          [ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83,  0, 1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83,  0,  1,  2])

def test_iter_sections_ipostorder():
    assert ([s.id for n in POP.neurites for s in n.iter_sections(Section.ipostorder)] ==
                          [1,  3,  5,  7,  9, 11, 13, 15, 17, 19, 20, 18, 16, 14, 12, 10,  8, 6,  4,  2,  0, 22, 24, 26, 28, 30, 32, 34, 36, 38, 40, 41, 39, 37, 35, 33, 31, 29, 27, 25, 23, 21, 43, 45, 47, 49, 51, 53, 55, 57, 59, 61, 62, 60, 58, 56, 54, 52, 50, 48, 46, 44, 42, 64, 66, 68, 70, 72, 74, 76, 78, 80, 82, 83, 81, 79, 77, 75, 73, 71, 69, 67, 65, 63,  1, 3,  5,  7,  9, 11, 13, 15, 17, 19, 20, 18, 16, 14, 12, 10,  8,  6, 4,  2,  0,  1,  3,  5,  7,  9, 11, 13, 15, 17, 19, 20, 18, 16, 14, 12, 10,  8,  6,  4,  2,  0, 22, 24, 26, 28, 30, 32, 34, 36, 38, 40, 41, 39, 37, 35, 33, 31, 29, 27, 25, 23, 21, 43, 45, 47, 49, 51, 53, 55, 57, 59, 61, 62, 60, 58, 56, 54, 52, 50, 48, 46, 44, 42, 64, 66, 68, 70, 72, 74, 76, 78, 80, 82, 83, 81, 79, 77, 75, 73, 71, 69, 67, 65, 63,  0,  1,  2])


def test_iter_sections_ibifurcation():
    assert ([s.id for n in POP.neurites for s in n.iter_sections(Section.ibifurcation_point)] ==
                          [0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 21, 23, 25, 27, 29, 31, 33, 35, 37, 39, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 63, 65, 67, 69, 71, 73, 75, 77, 79, 81, 0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 21, 23, 25, 27, 29, 31, 33, 35, 37, 39, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 63, 65, 67, 69, 71, 73, 75, 77, 79, 81])


def test_iter_sections_iforking():
    assert ([s.id for n in POP.neurites for s in n.iter_sections(Section.iforking_point)] ==
                          [0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 21, 23, 25, 27, 29, 31, 33, 35, 37, 39, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 63, 65, 67, 69, 71, 73, 75, 77, 79, 81, 0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 21, 23, 25, 27, 29, 31, 33, 35, 37, 39, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 63, 65, 67, 69, 71, 73, 75, 77, 79, 81])


def test_iter_sections_ileaf():
    assert ([s.id for n in POP.neurites for s in n.iter_sections(Section.ileaf)] ==
                          [1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 20, 22, 24, 26, 28, 30, 32, 34, 36, 38, 40, 41, 43, 45, 47, 49, 51, 53, 55, 57, 59, 61, 62, 64, 66, 68, 70, 72, 74, 76, 78, 80, 82, 83, 1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 20, 1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 20, 22, 24, 26, 28, 30, 32, 34, 36, 38, 40, 41, 43, 45, 47, 49, 51, 53, 55, 57, 59, 61, 62, 64, 66, 68, 70, 72, 74, 76, 78, 80, 82, 83, 0, 1, 2])


def test_iter_section_nrn():
    ref = list(iter_sections(SIMPLE))
    assert len(ref) == 6

    ref = list(iter_sections(SIMPLE, neurite_filter=lambda n: n.type == nm.AXON))
    assert len(ref) == 3

    ref = list(iter_sections(SIMPLE, neurite_filter=lambda n: n.type == nm.BASAL_DENDRITE))
    assert len(ref) == 3

    ref = list(iter_sections(SIMPLE, neurite_filter=lambda n: n.type == nm.APICAL_DENDRITE))
    assert len(ref) == 0


def test_iter_segments_nrn():
    ref = list(iter_segments(SIMPLE))
    assert len(ref) == 6

    ref = list(iter_segments(SIMPLE, neurite_filter=lambda n: n.type == nm.AXON))
    assert len(ref) == 3

    ref = list(iter_segments(SIMPLE, neurite_filter=lambda n: n.type == nm.BASAL_DENDRITE))
    assert len(ref) == 3

    ref = list(iter_segments(SIMPLE, neurite_filter=lambda n: n.type == nm.APICAL_DENDRITE))
    assert len(ref) == 0

    ref = list(iter_segments(NRN1))
    assert len(ref) == 840

    ref = list(iter_segments(NRN1, neurite_filter=lambda n: n.type == nm.AXON))
    assert len(ref) == 210

    ref = list(iter_segments(NRN1, neurite_filter=lambda n: n.type == nm.BASAL_DENDRITE))
    assert len(ref) == 420

    ref = list(iter_segments(NRN1, neurite_filter=lambda n: n.type == nm.APICAL_DENDRITE))
    assert len(ref) == 210


def test_iter_segments_pop():

    ref = list(iter_segments(POP))
    assert len(ref) == 3387

    ref = list(iter_segments(POP, neurite_filter=lambda n: n.type == nm.AXON))
    assert len(ref) == 919

    ref = list(iter_segments(POP, neurite_filter=lambda n: n.type == nm.BASAL_DENDRITE))
    assert len(ref) == 1549

    ref = list(iter_segments(POP, neurite_filter=lambda n: n.type == nm.APICAL_DENDRITE))
    assert len(ref) == 919


def test_iter_segments_section():
    sec = load_neuron(StringIO(u"""
	                      ((CellBody)
	                       (0 0 0 2))

	                      ((Dendrite)
                          (1 2 3 8)
                          (5 6 7 16)
                          (8 7 6 10)
                          (4 3 2 2))
                       """), reader='asc').sections[0]
    ref = [[p1[COLS.XYZR].tolist(), p2[COLS.XYZR].tolist()]
           for p1, p2 in iter_segments(sec)]

    assert_array_equal(ref, [[[1, 2, 3, 4], [5, 6, 7, 8]],
                             [[5, 6, 7, 8], [8, 7, 6, 5]],
                             [[8, 7, 6, 5], [4, 3, 2, 1]]])



#!/usr/bin/env python
# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Get sections and segments by ID."""

import neurom as nm
from neurom import morphmath as mm
from neurom.core.dataformat import COLS


def get_segment(neuron, section_id, segment_id):
    """Get a segment given a section and segment id

    Returns:
        array of two [x, y, z, r] points defining segment
    """
    sec = neuron.sections[section_id]
    return sec.points[segment_id:segment_id + 2][:, COLS.XYZR]


if __name__ == '__main__':

    nrn = nm.load_neuron('test_data/h5/v1/Neuron.h5')

    seg = get_segment(nrn, 3, 2)
    print('Segment:\n', seg)
    print('Mid-point (x, y, z):\n', mm.linear_interpolate(seg[0], seg[1], 0.5))
    print('Mid-point R:\n', mm.interpolate_radius(seg[0][COLS.R], seg[1][COLS.R], 0.5))
